"""
  This module contains methods for the validation of production definitions
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog


class ProdValidator(object):
    def __init__(self):
        self.transClient = TransformationClient()

    def checkTransStatus(self, transID):
        """Check if the status of the transformation is valid for the transformation to be added to a production.
            New is the only valid status

        :param int transID: the TransformationID
        """
        res = self.transClient.getTransformationParameters(transID, "Status")
        if not res["OK"]:
            return res
        status = res["Value"]
        if status != "New":
            return S_ERROR("checkTransStatus failed : Invalid transformation status: %s" % status)

        return S_OK()

    def checkTransDependency(self, transID, parentTransID):
        """Check if the transformation and the parent transformation are linked

        :param int transID: the TransformationID
        :param int parentTransID: the parent TransformationID
        """
        res = self.transClient.getTransformationMetaQuery(transID, "Input")
        if not res["OK"]:
            return res
        inputquery = res["Value"]
        if not inputquery:
            return S_ERROR("No InputMetaQuery defined for transformation %s" % transID)

        res = self.transClient.getTransformationMetaQuery(parentTransID, "Output")
        if not res["OK"]:
            return res
        parentoutputquery = res["Value"]
        if not parentoutputquery:
            return S_ERROR("No OutputMetaQuery defined for parent transformation %s" % parentTransID)

        # Check the matching between inputquery and parent outputmeta query
        # Currently very simplistic: just support expression with "=" and "in" operators
        gLogger.notice("Applying checkMatchQuery")
        res = self.checkMatchQuery(inputquery, parentoutputquery)

        if not res["OK"]:
            gLogger.error("checkMatchQuery failed")
            return res
        if not res["Value"]:
            return S_ERROR("checkMatchQuery result is False")

        return S_OK()

    def checkMatchQuery(self, mq, mqParent):
        """Check the logical intersection between the two metaqueries

        :param dict mq: a dictionary of the MetaQuery to be checked against the mqParent
        :param dict mqParent: a dictionary of the parent MetaQuery to be checked against the mq
        """
        # Get the metadata types defined in the catalog
        catalog = FileCatalog()
        res = catalog.getMetadataFields()
        if not res["OK"]:
            gLogger.error("Error in getMetadataFields: %s" % res["Message"])
            return res
        if not res["Value"]:
            gLogger.error("Error: no metadata fields defined")
            return res

        MetaTypeDict = res["Value"]["FileMetaFields"]
        MetaTypeDict.update(res["Value"]["DirectoryMetaFields"])

        res = self.checkformatQuery(mq)
        if not res["OK"]:
            return res
        MetaQueryDict = res["Value"]

        res = self.checkformatQuery(mqParent)
        if not res["OK"]:
            return res
        ParentMetaQueryDict = res["Value"]

        for meta, value in MetaQueryDict.items():
            if meta not in MetaTypeDict:
                msg = "Metadata %s is not defined in the Catalog" % meta
                return S_ERROR(msg)
            mtype = MetaTypeDict[meta]
            if mtype.lower() not in ["varchar(128)", "int", "float"]:
                msg = "Metatype %s is not supported" % mtype.lower()
                return S_ERROR(msg)
            if meta not in ParentMetaQueryDict:
                msg = "Metadata %s is not in parent transformation query" % meta
                return S_ERROR(msg)
            if self.compareValues(value, ParentMetaQueryDict[meta]):
                continue
            else:
                msg = "Metadata values %s do not match with %s" % (value, ParentMetaQueryDict[meta])
                gLogger.error(msg)
                return S_OK(False)

        return S_OK(True)

    def checkformatQuery(self, MetaQueryDict):
        """Check the format query and transform all dict values in dict for uniform treatment

        :param dict MetaQueryDict: a dictionary of the MetaQuery
        """
        for meta, value in MetaQueryDict.items():
            values = []
            if isinstance(value, dict):
                operation = list(value)[0]
                if operation not in ["=", "in"]:
                    msg = "Operation %s is not supported" % operation
                    return S_ERROR(msg)
                else:
                    if not isinstance(list(value.values())[0], list):
                        MetaQueryDict[meta] = {"in": list(value.values())}
            else:
                values.append(value)
                MetaQueryDict[meta] = {"in": values}

        return S_OK(MetaQueryDict)

    def compareValues(self, value, parentValue):
        """Very simple comparison. To be improved

        :param dict value: a dictionary with meta data values to be compared with the parentValues
        :param dict parentValue: a dictionary with meta data parentValues be compared with values
        """
        return set(list(value.values())[0]).issubset(set(list(parentValue.values())[0])) or set(
            list(parentValue.values())[0]
        ).issubset(set(list(value.values())[0]))



""" Callback when a staging operation is finished """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


from DIRAC import S_OK
from DIRAC.RequestManagementSystem.private.OperationHandlerBase import OperationHandlerBase
from DIRAC.WorkloadManagementSystem.Client.JobStateUpdateClient import JobStateUpdateClient


class StagingCallback(OperationHandlerBase):
    """
    .. class:: StagingCallback

       This performs the 'Done' callback to a job waiting for the staging to finish

       Currently, we cannot store the JobID in the field reserved in the Request, because
       then our crapy finalization system will try updating the job (minor) status
       So we store the job ID in the Argument field of operation
    """

    def __init__(self, operation=None, csPath=None):
        """c'tor

        :param Operation operation: an Operation instance
        :param str csPath: CS path for this handler
        """

        super(StagingCallback, self).__init__(operation, csPath)

    def __call__(self):
        """update the job status"""
        # # decode arguments
        jobID = self.operation.Arguments
        self.log.info("Performing callback to job %s" % jobID)

        res = JobStateUpdateClient().updateJobFromStager(jobID, "Done")

        if not res["OK"]:
            self.log.error("Error performing the callback to the job", res)
            return res

        self.operation.Status = "Done"
        self.log.info("Callback from staging done")
        return S_OK()



"""
It is a helper module which contains the available reports
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import hashlib
import re

from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.ConfigurationSystem.Client.PathFinder import getServiceSection
from DIRAC.MonitoringSystem.private.Plotters.BasePlotter import BasePlotter as myBasePlotter
from DIRAC.Core.Utilities.ObjectLoader import loadObjects

__RCSID__ = "$Id$"


class PlottersList(object):

    """
    Used to determine all available plotters used to create the plots

    :param dict __plotters: stores the available plotters
    """

    def __init__(self):
        """c'tor

        :param self: self reference
        """

        objectsLoaded = loadObjects(
            "MonitoringSystem/private/Plotters", re.compile(r".*[a-z1-9]Plotter\.py$"), myBasePlotter
        )
        self.__plotters = {}
        for objName in objectsLoaded:
            self.__plotters[objName[:-7]] = objectsLoaded[objName]

    def getPlotterClass(self, typeName):
        """
        It returns the plotter class for a given monitoring type
        """
        try:
            return S_OK(self.__plotters[typeName])
        except KeyError:
            return S_ERROR()


class MainReporter(object):
    """
    :param object __db: database object
    :param str __setup: DIRAC setup
    :param str __csSection: CS section used to configure some parameters.
    :param list __plotterList: available plotters
    """

    def __init__(self, db, setup):
        """c'tor

        :param self: self reference
        :param object: the database module
        :param str setup: DIRAC setup
        """
        self.__db = db
        self.__setup = setup
        self.__csSection = getServiceSection("Monitoring/Monitoring", setup=setup)
        self.__plotterList = PlottersList()

    def __calculateReportHash(self, reportRequest):
        """
        It creates an unique identifier

        :param dict reportRequest: plot attributes used to create the plot
        :return: an unique value
        :rtype: str
        """
        requestToHash = dict(reportRequest)
        granularity = gConfig.getValue("%s/CacheTimeGranularity" % self.__csSection, 300)
        for key in ("startTime", "endTime"):
            epoch = requestToHash[key]
            requestToHash[key] = epoch - epoch % granularity
        md5Hash = hashlib.md5()
        md5Hash.update(repr(requestToHash).encode())
        md5Hash.update(self.__setup.encode())
        return md5Hash.hexdigest()

    def generate(self, reportRequest, credDict):
        """
        It is used to create a plot.

        :param dict reportRequest: plot attributes used to create the plot

        .. note:: I know credDict is not used, but if we plan to add some policy, we need to use it!

        :return: dict S_OK/S_ERROR the values used to create the plot
        """
        typeName = reportRequest["typeName"]
        plotterClass = self.__plotterList.getPlotterClass(typeName)
        if not plotterClass["OK"]:
            return S_ERROR("There's no reporter registered for type %s" % typeName)

        reportRequest["hash"] = self.__calculateReportHash(reportRequest)
        plotter = plotterClass["Value"](self.__db, self.__setup, reportRequest["extraArgs"])
        return plotter.generate(reportRequest)

    def list(self, typeName):
        """
        It returns the available plots

        :param str typeName: monitoring type
        :return: dict S_OK/S_ERROR list of available reports (plots)
        """
        plotterClass = self.__plotterList.getPlotterClass(typeName)
        if not plotterClass["OK"]:
            return S_ERROR("There's no plotter registered for type %s" % typeName)
        plotter = plotterClass["Value"](self.__db, self.__setup)
        return S_OK(plotter.plotsList())



""" This object is a wrapper for setting and getting jobs states
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id"

import six
import datetime

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.WorkloadManagementSystem.Client.JobState.JobManifest import JobManifest
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB
from DIRAC.WorkloadManagementSystem.DB.JobLoggingDB import JobLoggingDB
from DIRAC.WorkloadManagementSystem.DB.TaskQueueDB import TaskQueueDB, singleValueDefFields, multiValueDefFields
from DIRAC.WorkloadManagementSystem.Service.JobPolicy import RIGHT_GET_INFO, RIGHT_RESCHEDULE
from DIRAC.WorkloadManagementSystem.Service.JobPolicy import RIGHT_RESET, RIGHT_CHANGE_STATUS


class JobState(object):
    class DBHold(object):
        def __init__(self):
            self.checked = False
            self.reset()

        def reset(self):
            self.jobDB = None
            self.logDB = None
            self.tqDB = None

    __db = DBHold()

    @classmethod
    def checkDBAccess(cls):
        # Init DB if there
        if not JobState.__db.checked:
            JobState.__db.jobDB = JobDB()
            JobState.__db.logDB = JobLoggingDB()
            JobState.__db.tqDB = TaskQueueDB()
            JobState.__db.checked = True

    def __init__(self, jid, source="Unknown"):
        self.__jid = jid
        self.__source = str(source)
        self.checkDBAccess()

    @property
    def jid(self):
        return self.__jid

    def setSource(self, source):
        self.__source = source

    def getManifest(self, rawData=False):
        result = JobState.__db.jobDB.getJobJDL(self.__jid)
        if not result["OK"] or rawData:
            return result
        if not result["Value"]:
            return S_ERROR("No manifest for job %s" % self.__jid)
        manifest = JobManifest()
        result = manifest.loadJDL(result["Value"])
        if not result["OK"]:
            return result
        return S_OK(manifest)

    def setManifest(self, manifest):
        if not isinstance(manifest, JobManifest):
            manifestStr = manifest
            manifest = JobManifest()
            result = manifest.load(manifestStr)
            if not result["OK"]:
                return result
        manifestJDL = manifest.dumpAsJDL()
        return self.__retryFunction(5, JobState.__db.jobDB.setJobJDL, (self.__jid, manifestJDL))

    # Execute traces

    def __retryFunction(self, retries, functor, args=False, kwargs=False):
        retries = max(1, retries)
        if not args:
            args = tuple()
        if not kwargs:
            kwargs = {}
        while retries:
            retries -= 1
            result = functor(*args, **kwargs)
            if result["OK"]:
                return result
            if retries == 0:
                return result
        return S_ERROR("No more retries")

    right_commitCache = RIGHT_GET_INFO

    def commitCache(self, initialState, cache, jobLog):
        try:
            self.__checkType(initialState, dict)
            self.__checkType(cache, dict)
            self.__checkType(jobLog, (list, tuple))
        except TypeError as excp:
            return S_ERROR(str(excp))
        result = self.getAttributes(list(initialState))
        if not result["OK"]:
            return result
        if not result["Value"] == initialState:
            return S_OK(False)
        gLogger.verbose("Job %s: About to execute trace. Current state %s" % (self.__jid, initialState))

        data = {"att": [], "jobp": [], "optp": []}
        for key in cache:
            for dk in data:
                if key.find("%s." % dk) == 0:
                    data[dk].append((key[len(dk) + 1 :], cache[key]))

        if data["att"]:
            attN = [t[0] for t in data["att"]]
            attV = [t[1] for t in data["att"]]
            result = self.__retryFunction(
                5, JobState.__db.jobDB.setJobAttributes, (self.__jid, attN, attV), {"update": True}
            )
            if not result["OK"]:
                return result

        if data["jobp"]:
            result = self.__retryFunction(5, JobState.__db.jobDB.setJobParameters, (self.__jid, data["jobp"]))
            if not result["OK"]:
                return result

        for k, v in data["optp"]:
            result = self.__retryFunction(5, JobState.__db.jobDB.setJobOptParameter, (self.__jid, k, v))
            if not result["OK"]:
                return result

        if "inputData" in cache:
            result = self.__retryFunction(5, JobState.__db.jobDB.setInputData, (self.__jid, cache["inputData"]))
            if not result["OK"]:
                return result

        gLogger.verbose("Adding logging records", " for %s" % self.__jid)
        for record, updateTime, source in jobLog:
            gLogger.verbose("", "Logging records for %s: %s %s %s" % (self.__jid, record, updateTime, source))
            record["date"] = updateTime
            record["source"] = source
            result = self.__retryFunction(5, JobState.__db.logDB.addLoggingRecord, (self.__jid,), record)
            if not result["OK"]:
                return result

        gLogger.info("Job %s: Ended trace execution" % self.__jid)
        # We return a new initial state
        return self.getAttributes(list(initialState))

    #
    # Status
    #

    def __checkType(self, value, tList, canBeNone=False):
        """Raise TypeError if the value does not have one of the expected types

        :param value: the value to test
        :param tList: type or tuple of types
        :param canBeNone: boolean, since there is no type for None to be used with isinstance

        """
        if canBeNone:
            if value is None:
                return
        if not isinstance(value, tList):
            raise TypeError("%s has wrong type. Has to be one of %s" % (value, tList))

    right_setStatus = RIGHT_GET_INFO

    def setStatus(self, majorStatus, minorStatus=None, appStatus=None, source=None, updateTime=None):
        try:
            self.__checkType(majorStatus, six.string_types)
            self.__checkType(minorStatus, six.string_types, canBeNone=True)
            self.__checkType(appStatus, six.string_types, canBeNone=True)
            self.__checkType(source, six.string_types, canBeNone=True)
            self.__checkType(updateTime, datetime.datetime, canBeNone=True)
        except TypeError as excp:
            return S_ERROR(str(excp))
        result = JobState.__db.jobDB.setJobStatus(
            self.__jid, status=majorStatus, minorStatus=minorStatus, applicationStatus=appStatus
        )
        if not result["OK"]:
            return result
        # HACK: Cause joblogging is crappy
        if not minorStatus:
            minorStatus = "idem"
        if not appStatus:
            appStatus = "idem"
        if not source:
            source = self.__source
        return JobState.__db.logDB.addLoggingRecord(
            self.__jid,
            status=majorStatus,
            minorStatus=minorStatus,
            applicationStatus=appStatus,
            date=updateTime,
            source=source,
        )

    right_getMinorStatus = RIGHT_GET_INFO

    def setMinorStatus(self, minorStatus, source=None, updateTime=None):
        try:
            self.__checkType(minorStatus, six.string_types)
            self.__checkType(source, six.string_types, canBeNone=True)
        except TypeError as excp:
            return S_ERROR(str(excp))
        result = JobState.__db.jobDB.setJobStatus(self.__jid, minorStatus=minorStatus)
        if not result["OK"]:
            return result
        if not source:
            source = self.__source
        return JobState.__db.logDB.addLoggingRecord(self.__jid, minorStatus=minorStatus, date=updateTime, source=source)

    def getStatus(self):
        result = JobState.__db.jobDB.getJobAttributes(self.__jid, ["Status", "MinorStatus"])
        if not result["OK"]:
            return result
        data = result["Value"]
        if data:
            return S_OK((data["Status"], data["MinorStatus"]))
        return S_ERROR("Job %d not found in the JobDB" % int(self.__jid))

    right_setAppStatus = RIGHT_GET_INFO

    def setAppStatus(self, appStatus, source=None, updateTime=None):
        try:
            self.__checkType(appStatus, six.string_types)
            self.__checkType(source, six.string_types, canBeNone=True)
        except TypeError as excp:
            return S_ERROR(str(excp))
        result = JobState.__db.jobDB.setJobStatus(self.__jid, applicationStatus=appStatus)
        if not result["OK"]:
            return result
        if not source:
            source = self.__source
        return JobState.__db.logDB.addLoggingRecord(
            self.__jid, applicationStatus=appStatus, date=updateTime, source=source
        )

    right_getAppStatus = RIGHT_GET_INFO

    def getAppStatus(self):
        result = JobState.__db.jobDB.getJobAttributes(self.__jid, ["ApplicationStatus"])
        if result["OK"]:
            result["Value"] = result["Value"]["ApplicationStatus"]
        return result

    # Attributes

    right_setAttribute = RIGHT_GET_INFO

    def setAttribute(self, name, value):
        try:
            self.__checkType(name, six.string_types)
            self.__checkType(value, six.string_types)
        except TypeError as excp:
            return S_ERROR(str(excp))
        return JobState.__db.jobDB.setJobAttribute(self.__jid, name, value)

    right_setAttributes = RIGHT_GET_INFO

    def setAttributes(self, attDict):
        try:
            self.__checkType(attDict, dict)
        except TypeError as excp:
            return S_ERROR(str(excp))
        keys = [key for key in attDict]
        values = [attDict[key] for key in keys]
        return JobState.__db.jobDB.setJobAttributes(self.__jid, keys, values)

    right_getAttribute = RIGHT_GET_INFO

    def getAttribute(self, name):
        try:
            self.__checkType(name, six.string_types)
        except TypeError as excp:
            return S_ERROR(str(excp))
        return JobState.__db.jobDB.getJobAttribute(self.__jid, name)

    right_getAttributes = RIGHT_GET_INFO

    def getAttributes(self, nameList=None):
        try:
            self.__checkType(nameList, (list, tuple), canBeNone=True)
        except TypeError as excp:
            return S_ERROR(str(excp))
        return JobState.__db.jobDB.getJobAttributes(self.__jid, nameList)

    # OptimizerParameters

    right_setOptParameter = RIGHT_GET_INFO

    def setOptParameter(self, name, value):
        try:
            self.__checkType(name, six.string_types)
            self.__checkType(value, six.string_types)
        except TypeError as excp:
            return S_ERROR(str(excp))
        return JobState.__db.jobDB.setJobOptParameter(self.__jid, name, value)

    right_setOptParameters = RIGHT_GET_INFO

    def setOptParameters(self, pDict):
        try:
            self.__checkType(pDict, dict)
        except TypeError as excp:
            return S_ERROR(str(excp))
        for name in pDict:
            result = JobState.__db.jobDB.setJobOptParameter(self.__jid, name, pDict[name])
            if not result["OK"]:
                return result
        return S_OK()

    right_removeOptParameters = RIGHT_GET_INFO

    def removeOptParameters(self, nameList):
        if isinstance(nameList, six.string_types):
            nameList = [nameList]
        try:
            self.__checkType(nameList, (list, tuple))
        except TypeError as excp:
            return S_ERROR(str(excp))
        for name in nameList:
            result = JobState.__db.jobDB.removeJobOptParameter(self.__jid, name)
            if not result["OK"]:
                return result
        return S_OK()

    right_getOptParameter = RIGHT_GET_INFO

    def getOptParameter(self, name):
        try:
            self.__checkType(name, six.string_types)
        except TypeError as excp:
            return S_ERROR(str(excp))
        return JobState.__db.jobDB.getJobOptParameter(self.__jid, name)

    right_getOptParameters = RIGHT_GET_INFO

    def getOptParameters(self, nameList=None):
        try:
            self.__checkType(nameList, (list, tuple), canBeNone=True)
        except TypeError as excp:
            return S_ERROR(str(excp))
        return JobState.__db.jobDB.getJobOptParameters(self.__jid, nameList)

    # Other

    right_resetJob = RIGHT_RESCHEDULE

    def rescheduleJob(self, source=""):
        result = JobState.__db.tqDB.deleteJob(self.__jid)
        if not result["OK"]:
            return S_ERROR("Cannot delete from TQ job %s: %s" % (self.__jid, result["Message"]))
        result = JobState.__db.jobDB.rescheduleJob(self.__jid)
        if not result["OK"]:
            return S_ERROR("Cannot reschedule in JobDB job %s: %s" % (self.__jid, result["Message"]))
        JobState.__db.logDB.addLoggingRecord(
            self.__jid, status=JobStatus.RECEIVED, minorStatus="", applicationStatus="", source=source
        )
        return S_OK()

    right_resetJob = RIGHT_RESET

    def resetJob(self, source=""):
        result = JobState.__db.jobDB.setJobAttribute(self.__jid, "RescheduleCounter", -1)
        if not result["OK"]:
            return S_ERROR("Cannot set the RescheduleCounter for job %s: %s" % (self.__jid, result["Message"]))
        result = JobState.__db.tqDB.deleteJob(self.__jid)
        if not result["OK"]:
            return S_ERROR("Cannot delete from TQ job %s: %s" % (self.__jid, result["Message"]))
        result = JobState.__db.jobDB.rescheduleJob(self.__jid)
        if not result["OK"]:
            return S_ERROR("Cannot reschedule in JobDB job %s: %s" % (self.__jid, result["Message"]))
        JobState.__db.logDB.addLoggingRecord(
            self.__jid, status=JobStatus.RECEIVED, minorStatus="", applicationStatus="", source=source
        )
        return S_OK()

    right_getInputData = RIGHT_GET_INFO

    def getInputData(self):
        return JobState.__db.jobDB.getInputData(self.__jid)

    @classmethod
    def checkInputDataStructure(cls, pDict):
        if not isinstance(pDict, dict):
            return S_ERROR("Input data has to be a dictionary")
        for lfn in pDict:
            if "Replicas" not in pDict[lfn]:
                return S_ERROR("Missing replicas for lfn %s" % lfn)
                replicas = pDict[lfn]["Replicas"]
                for seName in replicas:
                    if "SURL" not in replicas or "Disk" not in replicas:
                        return S_ERROR("Missing SURL or Disk for %s:%s replica" % (seName, lfn))
        return S_OK()

    right_setInputData = RIGHT_GET_INFO

    def set_InputData(self, lfnData):
        result = self.checkInputDataStructure(lfnData)
        if not result["OK"]:
            return result
        return JobState.__db.jobDB.setInputData(self.__jid, lfnData)

    right_insertIntoTQ = RIGHT_CHANGE_STATUS

    def insertIntoTQ(self, manifest=None):
        if not manifest:
            result = self.getManifest()
            if not result["OK"]:
                return result
            manifest = result["Value"]

        reqSection = "JobRequirements"

        result = manifest.getSection(reqSection)
        if not result["OK"]:
            return S_ERROR("No %s section in the job manifest" % reqSection)
        reqCfg = result["Value"]

        jobReqDict = {}
        for name in singleValueDefFields:
            if name in reqCfg:
                if name == "CPUTime":
                    jobReqDict[name] = int(reqCfg[name])
                else:
                    jobReqDict[name] = reqCfg[name]

        for name in multiValueDefFields:
            if name in reqCfg:
                jobReqDict[name] = reqCfg.getOption(name, [])

        jobPriority = reqCfg.getOption("UserPriority", 1)

        result = self.__retryFunction(2, JobState.__db.tqDB.insertJob, (self.__jid, jobReqDict, jobPriority))
        if not result["OK"]:
            errMsg = result["Message"]
            # Force removing the job from the TQ if it was actually inserted
            result = JobState.__db.tqDB.deleteJob(self.__jid)
            if result["OK"]:
                if result["Value"]:
                    gLogger.info("Job %s removed from the TQ" % self.__jid)
            return S_ERROR("Cannot insert in task queue: %s" % errMsg)
        return S_OK()



#! /usr/bin/env python
########################################################################
# File :    dirac-stager-monitor-request
# Author :  Andrew C. Smith
########################################################################
"""
Report the summary of the stage task from the DB.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("Request:  ID of the Stage request in the StorageManager")
    Script.parseCommandLine(ignoreErrors=False)

    args = Script.getPositionalArgs()

    if not len(args) == 1:
        Script.showHelp()

    from DIRAC import exit as DIRACExit, gLogger

    try:
        taskID = int(args[0])
    except Exception:
        gLogger.fatal("Stage requestID must be an integer")
        DIRACExit(2)

    from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient

    client = StorageManagerClient()

    res = client.getTaskSummary(taskID)
    if not res["OK"]:
        gLogger.error(res["Message"])
        DIRACExit(2)
    taskInfo = res["Value"]["TaskInfo"]
    replicaInfo = res["Value"]["ReplicaInfo"]
    outStr = "%s: %s" % ("TaskID".ljust(20), taskID)
    outStr += "\n%s: %s" % ("Status".ljust(20), taskInfo[taskID]["Status"])
    outStr += "\n%s: %s" % ("Source".ljust(20), taskInfo[taskID]["Source"])
    outStr += "\n%s: %s" % ("SourceTaskID".ljust(20), taskInfo[taskID]["SourceTaskID"])
    outStr += "\n%s: %s" % ("CallBackMethod".ljust(20), taskInfo[taskID]["CallBackMethod"])
    outStr += "\n%s: %s" % ("SubmitTime".ljust(20), taskInfo[taskID]["SubmitTime"])
    outStr += "\n%s: %s" % ("CompleteTime".ljust(20), taskInfo[taskID]["CompleteTime"])
    for lfn, metadata in replicaInfo.items():
        outStr += "\n"
        outStr += "\n\t%s: %s" % ("LFN".ljust(8), lfn.ljust(100))
        outStr += "\n\t%s: %s" % ("SE".ljust(8), metadata["StorageElement"].ljust(100))
        outStr += "\n\t%s: %s" % ("PFN".ljust(8), str(metadata["PFN"]).ljust(100))
        outStr += "\n\t%s: %s" % ("Size".ljust(8), str(metadata["FileSize"]).ljust(100))
        outStr += "\n\t%s: %s" % ("Status".ljust(8), metadata["Status"].ljust(100))
        outStr += "\n\t%s: %s" % ("Reason".ljust(8), str(metadata["Reason"]).ljust(100))
    gLogger.notice(outStr)


if __name__ == "__main__":
    main()



# FIXME: should be rewritten as real unittest

# __RCSID__ = "$Id$"
#
# import DIRAC
# from DIRAC.Core.Utilities.CFG import CFG
#
# DIRAC.gLogger.initialize('test_gConfig','/testSectionDebug')
#
# testconfig = '%s/DIRAC/ConfigurationSystem/test/test.cfg' % DIRAC.rootPath
# dumpconfig = '%s/DIRAC/ConfigurationSystem/test/dump.cfg' % DIRAC.rootPath
#
# cfg1 = CFG()
# cfg1.loadFromFile( testconfig )
#
# with open( testconfig ) as fd:
#   cfg1String = fd.read()
#
# cfg2 = CFG()
# cfg2.loadFromBuffer( cfg1.serialize() )
#
# cfg3 = cfg1.mergeWith( cfg2 )
#
# testList = [{ 'method'    : DIRAC.gConfig.loadFile,
#               'arguments' : ( testconfig, ),
#               'output'    : {'OK': True, 'Value': ''}
#             },
#             { 'method'    : DIRAC.gConfig.dumpLocalCFGToFile,
#               'arguments' : ( dumpconfig, ),
#               'output'    : {'OK': True, 'Value': ''}
#             },
#             { 'method'    : cfg1.serialize,
#               'arguments' : ( ),
#               'output'    : cfg1String
#             },
#             { 'method'    : cfg3.serialize,
#               'arguments' : ( ),
#               'output'    : cfg1String
#             }]
#
# testdict = { 'DIRAC.gConfig'               : testList,}
#
# DIRAC.Tests.run( testdict, 'DIRAC.gConfig.files' )
#
#
#
# # testList = [{ 'method'    : DIRAC.gConfig.get,
# #               'arguments' : ( '/testSection/test', ),
# #               'output'    : {'OK': True, 'Value': 'test'}
# #             },
# # #             { 'method'    : DIRAC.gConfig.get,
# # #              'arguments' : ( '/testSection/nonexisting','OK', ),
# # #              'output'    : {'OK': True, 'Value': 'test'}
# # #            },
# #              ]
# #
# # testdict = { 'DIRAC.gConfig'               : testList,}
# #
# #
# # DIRAC.Tests.run( testdict, 'DIRAC.gConfig' )



""".. module:: Test_DAG

Test cases for DIRAC.Core.Utilities.DAG module.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import unittest

# sut
from DIRAC.Core.Utilities.DAG import DAG, makeFrozenSet

__RCSID__ = "$Id $"


def listToSet(obj, recursive=False):
    """Convert a list of objects into a set which can be used for comparison"""
    result = set()
    for x in obj:
        if isinstance(x, (list, tuple)):
            x = frozenset(x) if recursive else tuple(x)
        elif isinstance(x, dict):
            x = tuple(map(tuple, sorted(x.items())))
        result.add(x)
    return result


########################################################################
class DAGTestCase(unittest.TestCase):
    """Test case for DIRAC.Core.Utilities.DAG module"""

    pass


class DAGSimple(DAGTestCase):
    def test_makeFrozenSet(self):
        """test makeFrozenSet"""
        res = makeFrozenSet({"a": "b"})
        self.assertEqual(res, frozenset({("a", "b")}))

        # dict with lists in
        dList1 = {"a": []}
        res = makeFrozenSet(dList1)
        self.assertEqual(res, frozenset({("a", frozenset([]))}))

        dList2 = {"a": [0, 1]}
        res = makeFrozenSet(dList2)
        self.assertEqual(res, frozenset({("a", frozenset([0, 1]))}))

        dList3 = {"a": [0, 1], "b": 0}
        res = makeFrozenSet(dList3)
        self.assertEqual(res, frozenset({("a", frozenset([0, 1])), ("b", 0)}))

        # dict with sets in
        dSet1 = {"a": set()}
        res = makeFrozenSet(dSet1)
        self.assertEqual(res, frozenset({("a", frozenset([]))}))

        # dict with dicts in
        dDict1 = {"a": {"a": "b"}}
        res = makeFrozenSet(dDict1)
        self.assertEqual(res, frozenset({("a", frozenset({("a", "b")}))}))

        # #dicts with sets, list, and dicts in
        dAll = {"a": {"a": "b"}, "c": [0, 1], "d": set()}
        res = makeFrozenSet(dAll)
        self.assertEqual(
            res, frozenset({("a", frozenset({("a", "b")})), ("c", frozenset([0, 1])), ("d", frozenset([]))})
        )


class DAGFull(DAGTestCase):
    def test_getList(self):
        """test dag to list"""
        dag = DAG()
        dag.addNode("A")
        l = dag.getList()
        self.assertEqual(l, ["A"])

        dag.addNode("C")
        dag.addEdge("A", "C")
        l = dag.getList()
        self.assertEqual(l, ["A", "C"])

        dag.addNode("B")
        dag.addEdge("C", "B")
        l = dag.getList()
        self.assertEqual(l, ["A", "C", "B"])

        d = dict(zip("ab", range(2)))
        dag.addNode(d)
        dag.addEdge("B", d)
        l = dag.getList()
        self.assertEqual(l, ["A", "C", "B", d])

        l1 = list(range(2))
        dag.addNode(l1)
        dag.addEdge(d, l1)
        l = dag.getList()
        self.assertEqual(l, ["A", "C", "B", d, l1])

        dag.addNode("E")
        dag.addEdge(l1, "E")
        l = dag.getList()
        self.assertEqual(l, ["A", "C", "B", d, l1, "E"])

        dag1 = DAG()
        dag1.addNode(d)
        dag1.addNode(l1)
        dag1.addEdge(d, l1)
        l = dag1.getList()
        self.assertEqual(l, [d, l1])

    def test_full(self):
        """test dag creation and more"""
        dag = DAG()
        i_n = dag.getIndexNodes()
        self.assertEqual(i_n, [])
        dag.addNode("A")
        self.assertEqual(dag.graph, {"A": set()})
        dag.addNode("A")
        self.assertEqual(dag.graph, {"A": set()})
        dag.addNode("B")
        self.assertEqual(dag.graph, {"A": set(), "B": set()})
        l = dag.getList()
        self.assertEqual(l, [])
        dag.addEdge("A", "B")
        self.assertEqual(dag.graph, {"A": {"B"}, "B": set()})
        l = dag.getList()
        self.assertEqual(set(l), {"A", "B"})
        dag.addEdge("A", "B")
        self.assertEqual(dag.graph, {"A": {"B"}, "B": set()})
        dag.addEdge("A", "C")
        self.assertEqual(dag.graph, {"A": {"B"}, "B": set()})
        dag.addNode("C")
        dag.addEdge("A", "C")
        self.assertEqual(dag.graph, {"A": {"B", "C"}, "B": set(), "C": set()})
        l = dag.getList()
        self.assertEqual(l, ["A"])
        dag.addEdge("C", "A")  # this would be cyclic, so it should not change the graph
        self.assertEqual(dag.graph, {"A": {"B", "C"}, "B": set(), "C": set()})
        dag.addNode("D")
        i_n = dag.getIndexNodes()
        self.assertEqual(set(i_n), {"A", "D"})
        dag.addNode("E")
        i_n = dag.getIndexNodes()
        self.assertEqual(listToSet(i_n), listToSet(["A", "D", "E"]))
        dag.addEdge("A", "D")
        dag.addEdge("D", "E")
        self.assertEqual(dag.graph, {"A": {"B", "C", "D"}, "B": set(), "C": set(), "D": {"E"}, "E": set()})
        i_n = dag.getIndexNodes()
        self.assertEqual(i_n, ["A"])
        dag.addEdge("E", "A")
        self.assertEqual(dag.graph, {"A": {"B", "C", "D"}, "B": set(), "C": set(), "D": {"E"}, "E": {"A"}})

        # now an object
        class forTest(object):
            pass

        ft = forTest()
        dag.addNode(ft)
        self.assertEqual(dag.graph, {"A": {"B", "C", "D"}, "B": set(), "C": set(), "D": {"E"}, "E": {"A"}, ft: set()})
        dag.addEdge("B", ft)
        self.assertEqual(dag.graph, {"A": {"B", "C", "D"}, "B": {ft}, "C": set(), "D": {"E"}, "E": {"A"}, ft: set()})

        # now sets, dicts and lists as nodes
        d = dict(zip("ab", range(2)))
        dag.addNode(d)
        self.assertEqual(
            dag.graph,
            {
                "A": {"B", "C", "D"},
                "B": {ft},
                "C": set(),
                "D": {"E"},
                "E": {"A"},
                ft: set(),
                frozenset({("a", 0), ("b", 1)}): set(),
            },
        )
        dag.addEdge(ft, d)
        self.assertEqual(
            dag.graph,
            {
                "A": {"B", "C", "D"},
                "B": {ft},
                "C": set(),
                "D": {"E"},
                "E": {"A"},
                ft: set([frozenset({("a", 0), ("b", 1)})]),
                frozenset({("a", 0), ("b", 1)}): set(),
            },
        )

        l = list(range(2))
        dag.addNode(l)
        self.assertEqual(
            dag.graph,
            {
                "A": {"B", "C", "D"},
                "B": {ft},
                "C": set(),
                "D": {"E"},
                "E": {"A"},
                ft: set([frozenset({("a", 0), ("b", 1)})]),  # ft -> d
                frozenset({("a", 0), ("b", 1)}): set(),  # d
                frozenset({0, 1}): set(),  # l
            },
        )
        dag.addEdge(d, l)
        self.assertEqual(
            dag.graph,
            {
                "A": {"B", "C", "D"},
                "B": {ft},
                "C": set(),
                "D": {"E"},
                "E": {"A"},
                ft: set([frozenset({("a", 0), ("b", 1)})]),  # ft -> d
                frozenset({("a", 0), ("b", 1)}): set([frozenset({0, 1})]),  # d->l
                frozenset({0, 1}): set(),  # l
            },
        )

        del dag.graph["E"]
        del dag.graph["D"]
        del dag.graph[ft]
        del dag.graph[frozenset({("a", 0), ("b", 1)})]
        dag.graph["A"] = {"B", "C"}

        self.assertEqual(
            dag.graph,
            {
                "A": {"B", "C"},
                "B": {ft},
                "C": set(),
                frozenset({0, 1}): set(),  # l
            },
        )

        i_n = dag.getIndexNodes()
        self.assertEqual(listToSet(i_n), listToSet(["A", l]))

        d1 = {"a": "b"}
        dag.addNode(d1)
        self.assertEqual(
            dag.graph,
            {
                "A": {"B", "C"},
                "B": {ft},
                "C": set(),
                frozenset({0, 1}): set(),  # l
                frozenset({("a", "b")}): set(),  # d1
            },
        )

        l1 = ["a", "b"]
        dag.addNode(l1)
        self.assertEqual(
            dag.graph,
            {
                "A": {"B", "C"},
                "B": {ft},
                "C": set(),
                frozenset({0, 1}): set(),  # l
                frozenset({("a", "b")}): set(),  # d1
                frozenset({"a", "b"}): set(),  # l1
            },
        )

        i_n = dag.getIndexNodes()
        self.assertEqual(listToSet(i_n, recursive=True), listToSet(["A", l, d1, l1], recursive=True))

        s1 = set()
        dag.addNode(s1)
        self.assertEqual(
            dag.graph,
            {
                "A": {"B", "C"},
                "B": {ft},
                "C": set(),
                frozenset({0, 1}): set(),  # l
                frozenset({("a", "b")}): set(),  # d1
                frozenset({"a", "b"}): set(),  # l1
                frozenset({}): set(),  # s1
            },
        )

        # dict with frozenset in
        dFSet1 = {"a": frozenset()}
        dag.addNode(dFSet1)
        self.assertEqual(
            dag.graph,
            {
                "A": {"B", "C"},
                "B": {ft},
                "C": set(),
                frozenset({0, 1}): set(),  # l
                frozenset({("a", "b")}): set(),  # d1
                frozenset({"a", "b"}): set(),  # l1
                frozenset({}): set(),  # s1
                frozenset([("a", frozenset([]))]): set(),  # dFSet1
            },
        )


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(DAGTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(DAGSimple))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(DAGFull))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



""" JobLoggingDB class is a front-end to the Job Logging Database.
    The following methods are provided

    addLoggingRecord()
    getJobLoggingInfo()
    deleteJob()
    getWMSTimeStamps()
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time
import six

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Base.DB import DB

MAGIC_EPOC_NUMBER = 1270000000

#############################################################################


class JobLoggingDB(DB):
    """Frontend to JobLoggingDB MySQL table"""

    def __init__(self):
        """Standard Constructor"""

        DB.__init__(self, "JobLoggingDB", "WorkloadManagement/JobLoggingDB")

    #############################################################################
    def addLoggingRecord(
        self,
        jobID,
        status=None,
        minorStatus=None,
        applicationStatus=None,
        date=None,
        source=None,
        minor=None,
        application=None,
    ):
        """Add a new entry to the JobLoggingDB table. One, two or all the three status
        components (status, minorStatus, applicationStatus) can be specified.
        Optionally the time stamp of the status can
        be provided in a form of a string in a format '%Y-%m-%d %H:%M:%S' or
        as datetime.datetime object. If the time stamp is not provided the current
        UTC time is used.
        """

        # Backward compatibility
        # FIXME: to remove in next version
        if minor:
            minorStatus = minor
        if application:
            applicationStatus = application

        status = status or "idem"
        minorStatus = minorStatus or "idem"
        applicationStatus = applicationStatus or "idem"
        source = source or "Unknown"

        event = "status/minor/app=%s/%s/%s" % (status, minorStatus, applicationStatus)
        self.log.info("Adding record for job ", str(jobID) + ": '" + event + "' from " + source)

        try:
            if not date:
                # Make the UTC datetime string and float
                _date = Time.dateTime()
            elif isinstance(date, six.string_types):
                # The date is provided as a string in UTC
                _date = Time.fromString(date)
            elif isinstance(date, Time._dateTimeType):
                _date = date
            else:
                self.log.error("Incorrect date for the logging record")
                _date = Time.dateTime()
        except Exception:
            self.log.exception("Exception while date evaluation")
            _date = Time.dateTime()
        epoc = time.mktime(_date.timetuple()) + _date.microsecond / 1000000.0 - MAGIC_EPOC_NUMBER

        cmd = (
            "INSERT INTO LoggingInfo (JobId, Status, MinorStatus, ApplicationStatus, "
            + "StatusTime, StatusTimeOrder, StatusSource) VALUES (%d,'%s','%s','%s','%s',%f,'%s')"
            % (int(jobID), status, minorStatus, applicationStatus[:255], str(_date), epoc, source[:32])
        )

        return self._update(cmd)

    #############################################################################
    def getJobLoggingInfo(self, jobID):
        """Returns a Status,MinorStatus,ApplicationStatus,StatusTime,StatusSource tuple
        for each record found for job specified by its jobID in historical order
        """

        cmd = (
            "SELECT Status,MinorStatus,ApplicationStatus,StatusTime,StatusSource FROM"
            " LoggingInfo WHERE JobId=%d ORDER BY StatusTimeOrder,StatusTime" % int(jobID)
        )

        result = self._query(cmd)
        if not result["OK"]:
            return result
        if result["OK"] and not result["Value"]:
            return S_ERROR("No Logging information for job %d" % int(jobID))

        return_value = []
        status, minor, app = result["Value"][0][:3]
        if app == "idem":
            app = "Unknown"
        for row in result["Value"]:
            if row[0] != "idem":
                status = row[0]
            if row[1] != "idem":
                minor = row[1]
            if row[2] != "idem":
                app = row[2]
            return_value.append((status, minor, app, str(row[3]), row[4]))

        return S_OK(return_value)

    #############################################################################
    def deleteJob(self, jobID):
        """Delete logging records for given jobs"""
        if not jobID:
            return S_OK()

        # Make sure that we have a list of strings of jobIDs
        if isinstance(jobID, six.integer_types):
            jobList = [str(jobID)]
        elif isinstance(jobID, six.string_types):
            jobList = jobID.replace(" ", "").split(",")
        else:
            jobList = list(str(j) for j in jobID)

        req = "DELETE FROM LoggingInfo WHERE JobID IN (%s)" % ",".join(jobList)
        return self._update(req)

    #############################################################################
    def getWMSTimeStamps(self, jobID):
        """Get TimeStamps for job MajorState transitions
        return a {State:timestamp} dictionary
        """
        # self.log.debug('getWMSTimeStamps: Retrieving Timestamps for Job %d' % int(jobID))

        result = {}
        cmd = "SELECT Status,StatusTimeOrder FROM LoggingInfo WHERE JobID=%d" % int(jobID)
        resCmd = self._query(cmd)
        if not resCmd["OK"]:
            return resCmd
        if not resCmd["Value"]:
            return S_ERROR("No Logging Info for job %d" % int(jobID))

        for event, etime in resCmd["Value"]:
            result[event] = str(etime + MAGIC_EPOC_NUMBER)

        # Get last date and time
        cmd = "SELECT MAX(StatusTime) FROM LoggingInfo WHERE JobID=%d" % int(jobID)
        resCmd = self._query(cmd)
        if not resCmd["OK"]:
            return resCmd
        if resCmd["Value"]:
            result["LastTime"] = str(resCmd["Value"][0][0])
        else:
            result["LastTime"] = "Unknown"

        return S_OK(result)



"""
  Defines the plugin to take storage space information given by BDII
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC import S_OK
from DIRAC.Core.Utilities.Grid import ldapsearchBDII


class BDIIOccupancy(object):
    """.. class:: BDIIOccupancy

    Occupancy plugin to return the space information given by BDII
    Assuming the protocol is SRM
    """

    def __init__(self, se):
        # flag to show initalization status of the plugin
        self.log = se.log.getSubLogger("BDIIOccupancy")
        # BDII host to query
        self.bdii = "cclcgtopbdii01.in2p3.fr:2170"
        if "LCG_GFAL_INFOSYS" in os.environ:
            self.bdii = os.environ["LCG_GFAL_INFOSYS"]
        self.vo = se.vo
        # assume given SE speaks SRM
        ret = se.getStorageParameters(protocol="srm")
        if not ret["OK"]:
            raise RuntimeError(ret["Message"])
        if "Host" not in ret["Value"]:
            raise RuntimeError("No Host is found from StorageParameters")
        self.host = ret["Value"]["Host"]

    def getOccupancy(self, **kwargs):
        """Returns the space information given by BDII
        Total and Free space are taken from GlueSATotalOnlineSize and GlueSAFreeOnlineSize, respectively.

        :returns: S_OK with dict (keys: Total, Free)
        """
        sTokenDict = {"Total": 0, "Free": 0}
        BDIIAttr = ["GlueSATotalOnlineSize", "GlueSAFreeOnlineSize"]

        filt = "(&(GlueSAAccessControlBaseRule=VO:%s)(GlueChunkKey=GlueSEUniqueID=%s))" % (self.vo, self.host)
        ret = ldapsearchBDII(filt, BDIIAttr, host=self.bdii)
        if not ret["OK"]:
            return ret
        for value in ret["Value"]:
            if "attr" in value:
                attr = value["attr"]
                sTokenDict["Total"] = float(attr.get(BDIIAttr[0], 0)) * 1024 * 1024 * 1024
                sTokenDict["Free"] = float(attr.get(BDIIAttr[1], 0)) * 1024 * 1024 * 1024
        return S_OK(sTokenDict)



########################################################################
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2011/12/14 15:07:01
########################################################################

""" :mod: PfnTestCase
    =======================

    .. module: PfnTestCase
    :synopsis: test case for Pfn module
    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    test case for Pfn module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

# #
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2011/12/14 15:07:12

# # imports
import unittest

# sut
from DIRAC.Core.Utilities.Pfn import pfnparse, pfnunparse


########################################################################
class PfnTests(unittest.TestCase):
    """
    .. class:: PfnTests

    """

    def setUp(self):
        self.default_pfns = {
            None: {
                "Errno": 0,
                "Message": "wrong 'pfn' argument value in function call, expected non-empty string, got <type 'NoneType'>",
                "OK": False,
            },
            "": {
                "OK": False,
                "Errno": 0,
                "Message": "wrong 'pfn' argument value in function call, expected non-empty string, got <type 'NoneType'>",
            },
            "/a/b/c": {
                "OK": True,
                "Value": {"Protocol": "", "WSUrl": "", "FileName": "c", "Host": "", "Path": "/a/b", "Port": ""},
            },
            "proto:/a/b/c": {
                "OK": True,
                "Value": {"Protocol": "proto", "WSUrl": "", "FileName": "c", "Host": "", "Path": "/a/b", "Port": ""},
            },
            "proto://host/a/b/c": {
                "OK": True,
                "Value": {
                    "Protocol": "proto",
                    "WSUrl": "",
                    "FileName": "c",
                    "Host": "host",
                    "Path": "/a/b",
                    "Port": "",
                },
            },
            "proto://host:port/a/b/c": {
                "OK": True,
                "Value": {
                    "Protocol": "proto",
                    "WSUrl": "",
                    "FileName": "c",
                    "Host": "host",
                    "Path": "/a/b",
                    "Port": "port",
                },
            },
            "proto://host:port//a/b/c?SvcClass=toto": {
                "OK": True,
                "Value": {
                    "Protocol": "proto",
                    "WSUrl": "",
                    "FileName": "c",
                    "Host": "host",
                    "Path": "//a/b",
                    "Port": "port",
                    "Options": "SvcClass=toto",
                },
            },
            "proto://host:port/a/b/c?SvcClass=toto": {
                "OK": True,
                "Value": {
                    "Protocol": "proto",
                    "WSUrl": "",
                    "FileName": "c",
                    "Host": "host",
                    "Path": "/a/b",
                    "Port": "port",
                    "Options": "SvcClass=toto",
                },
            },
        }

        # We keep some standard non srm specific url that used to be supported, just for backward compatibility
        self.srm_pfns = {
            None: {
                "Errno": 0,
                "Message": "wrong 'pfn' argument value in function call, expected non-empty string, got <type 'NoneType'>",
                "OK": False,
            },
            "": {
                "OK": False,
                "Errno": 0,
                "Message": "wrong 'pfn' argument value in function call, expected non-empty string, got <type 'NoneType'>",
            },
            "/a/b/c": {
                "OK": True,
                "Value": {"Protocol": "", "WSUrl": "", "FileName": "c", "Host": "", "Path": "/a/b", "Port": ""},
            },
            "proto:/a/b/c": {
                "OK": True,
                "Value": {"Protocol": "proto", "WSUrl": "", "FileName": "c", "Host": "", "Path": "/a/b", "Port": ""},
            },
            "proto://host/a/b/c": {
                "OK": True,
                "Value": {
                    "Protocol": "proto",
                    "WSUrl": "",
                    "FileName": "c",
                    "Host": "host",
                    "Path": "/a/b",
                    "Port": "",
                },
            },
            "proto://host:port/a/b/c": {
                "OK": True,
                "Value": {
                    "Protocol": "proto",
                    "WSUrl": "",
                    "FileName": "c",
                    "Host": "host",
                    "Path": "/a/b",
                    "Port": "port",
                },
            },
            "proto://host:port/wsurl?=/a/b/c": {
                "OK": True,
                "Value": {
                    "Protocol": "proto",
                    "WSUrl": "/wsurl?=",
                    "FileName": "c",
                    "Host": "host",
                    "Path": "/a/b",
                    "Port": "port",
                },
            },
            "proto://host:port/wsurl?blah=/a/b/c": {
                "OK": True,
                "Value": {
                    "Protocol": "proto",
                    "WSUrl": "/wsurl?blah=",
                    "FileName": "c",
                    "Host": "host",
                    "Path": "/a/b",
                    "Port": "port",
                },
            },
        }

    def test_01_srm_parse(self):
        """pfnparse and pfnparse_old

        :param self: self reference
        """

        for pfn, result in self.srm_pfns.items():
            parseResult = pfnparse(pfn)
            self.assertEqual(parseResult["OK"], result["OK"])
            if result["OK"]:
                self.assertEqual(parseResult["Value"], result["Value"])

    def test_02_default_parse(self):
        """pfnparse and pfnparse_old

        :param self: self reference
        """

        for pfn, result in self.default_pfns.items():
            parseResult = pfnparse(pfn, srmSpecific=False)
            self.assertEqual(parseResult["OK"], result["OK"])
            if result["OK"]:
                self.assertEqual(parseResult["Value"], result["Value"])

    def test_03_srm_unparse(self):
        """pfnunparse and pfnunparse_old

        :param self: self reference
        """
        for pfn, result in self.srm_pfns.items():
            if result["OK"]:
                unparseResult = pfnunparse(result["Value"])
                self.assertEqual(unparseResult, {"OK": True, "Value": pfn})
        self.assertEqual(pfnunparse(None)["OK"], False)
        self.assertEqual(pfnunparse("Path")["OK"], False)

    def test_03_default_unparse(self):
        """pfnunparse and pfnunparse_old

        :param self: self reference
        """
        for pfn, result in self.default_pfns.items():
            if result["OK"]:
                unparseResult = pfnunparse(result["Value"], srmSpecific=False)
                self.assertEqual(unparseResult, {"OK": True, "Value": pfn})
        self.assertEqual(pfnunparse(None)["OK"], False)
        self.assertEqual(pfnunparse("Path")["OK"], False)


# # test execution
if __name__ == "__main__":
    testLoader = unittest.TestLoader()
    suite = testLoader.loadTestsFromTestCase(PfnTests)
    unittest.TextTestRunner(verbosity=3).run(suite)



import json

from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation

from DIRAC.TransformationSystem.Client.BodyPlugin.BaseBody import BaseBody


class DummyBody(BaseBody):
    """Dummy BodyPlugin class that just creates
    a ForwardDISET operation, whose argument
    is the number of LFNs multiplied by a factor
    It is used mostly as an example and for unit tests.
    """

    # This is needed to know how to serialize the object
    _attrToSerialize = ["factor"]

    def __init__(self, factor=0):
        """C'tor

        :param factor: multiplying factor
        """
        self.factor = factor

    def taskToRequest(self, taskID, task, transID):
        """Convert a task into an RMS with a single ForwardDISET
        Operation whose attribute is the number of files in the task
        """
        req = Request()
        op = Operation()
        op.Type = "ForwardDISET"
        op.Arguments = json.dumps(len(task["InputData"]) * self.factor)
        req.addOperation(op)
        return req



########################################################################
# File :   AgentReactor.py
# Author : Adria Casajus
########################################################################
"""
  DIRAC class to execute Agents

  Agents are the active part any any DIRAC system, they execute in a cyclic
  manner looking at the state of the system and reacting to it by taken
  appropriated actions

  All DIRAC Agents must inherit from the basic class AgentModule

  In the most common case, DIRAC Agents are executed using the dirac-agent command.
  dirac-agent accepts a list positional arguments. These arguments have the form:
  [DIRAC System Name]/[DIRAC Agent Name]
  dirac-agent then:
  - produces a instance of AgentReactor
  - loads the required modules using the AgentReactor.loadAgentModules method
  - starts the execution loop using the AgentReactor.go method

  Agent modules must be placed under the Agent directory of a DIRAC System.
  DIRAC Systems are called XXXSystem where XXX is the [DIRAC System Name], and
  must inherit from the base class AgentModule

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import time

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.Base.private.ModuleLoader import ModuleLoader
from DIRAC.Core.Utilities import ThreadScheduler
from DIRAC.Core.Base.AgentModule import AgentModule

__RCSID__ = "$Id$"


class AgentReactor(object):
    """
    Main interface to DIRAC Agents. It allows to :
    - define a Agents modules to be executed
    - define the number of cycles to execute
    - steer the execution

    Agents are declared via:
    - loadAgentModule(): for a single Agent
    - loadAgentModules(): for a list of Agents

    The number of cycles to execute for a defined Agent can be set via:
    - setAgentModuleCyclesToExecute()

    The execution of the Agents is done with:
    - runNumCycles(): to execute an additional number of cycles
    - go():

    During the execution of the cycles, each of the Agents can be signaled to stop
    by creating a file named "stop_agent" in its Control Directory.

    """

    def __init__(self, baseAgentName):
        self.__agentModules = {}
        self.__loader = ModuleLoader("Agent", PathFinder.getAgentSection, AgentModule)
        self.__tasks = {}
        self.__baseAgentName = baseAgentName
        self.__scheduler = ThreadScheduler.ThreadScheduler(enableReactorThread=False, minPeriod=30)
        self.__alive = True
        self.__running = False

    def loadAgentModules(self, modulesList, hideExceptions=False):
        """
        Load all modules required in moduleList
        """
        result = self.__loader.loadModules(modulesList, hideExceptions=hideExceptions)
        if not result["OK"]:
            return result
        self.__agentModules = self.__loader.getModules()
        for agentName in self.__agentModules:
            agentData = self.__agentModules[agentName]
            agentData["running"] = False
            try:
                instanceObj = agentData["classObj"](agentName, agentData["loadName"], self.__baseAgentName)
                result = instanceObj.am_initialize()
                if not result["OK"]:
                    return S_ERROR("Error while calling initialize method of %s: %s" % (agentName, result["Message"]))
                agentData["instanceObj"] = instanceObj
            except Exception as excp:
                if not hideExceptions:
                    gLogger.exception("Can't load agent %s" % agentName, lException=excp)
                return S_ERROR("Can't load agent %s: \n %s" % (agentName, excp))
            agentPeriod = instanceObj.am_getPollingTime()
            result = self.__scheduler.addPeriodicTask(
                agentPeriod, instanceObj.am_go, executions=instanceObj.am_getMaxCycles(), elapsedTime=agentPeriod
            )
            if not result["OK"]:
                return result

            taskId = result["Value"]
            self.__tasks[result["Value"]] = agentName
            agentData["taskId"] = taskId
            agentData["running"] = True

        if not self.__agentModules:
            return S_ERROR("No agent module loaded")

        return S_OK()

    def runNumCycles(self, agentName=None, numCycles=1):
        """
        Run all defined agents a given number of cycles
        """
        if agentName:
            self.loadAgentModules([agentName])
        error = ""
        for aName in self.__agentModules:
            result = self.setAgentModuleCyclesToExecute(aName, numCycles)
            if not result["OK"]:
                error = "Failed to set cycles to execute"
                gLogger.error("%s:" % error, aName)
                break
        if error:
            return S_ERROR(error)
        self.go()
        return S_OK()

    def __finalize(self):
        """
        Execute the finalize method of all Agents
        """
        for agentName in self.__agentModules:
            try:
                self.__agentModules[agentName]["instanceObj"].finalize()
            except Exception as excp:
                gLogger.exception("Failed to execute finalize for Agent: %s" % agentName, lException=excp)

    def go(self):
        """
        Main method to control the execution of all configured Agents
        """
        if self.__running:
            return
        self.__running = True
        try:
            while self.__alive:
                self.__checkControlDir()
                timeToNext = self.__scheduler.executeNextTask()
                if timeToNext is None:
                    gLogger.info("No more agent modules to execute. Exiting")
                    break
                time.sleep(min(max(timeToNext, 0.5), 5))
        finally:
            self.__running = False
        self.__finalize()

    def setAgentModuleCyclesToExecute(self, agentName, maxCycles=1):
        """
        Set number of cycles to execute for a given agent (previously defined)
        """
        if agentName not in self.__agentModules:
            return S_ERROR("%s has not been loaded" % agentName)
        if maxCycles:
            try:
                maxCycles += self.__agentModules[agentName]["instanceObj"].am_getCyclesDone()
            except Exception as excp:
                error = "Can not determine number of cycles to execute"
                gLogger.exception("%s: '%s'" % (error, maxCycles), lException=excp)
                return S_ERROR(error)
        self.__agentModules[agentName]["instanceObj"].am_setOption("MaxCycles", maxCycles)
        self.__scheduler.setNumExecutionsForTask(self.__agentModules[agentName]["taskId"], maxCycles)
        return S_OK()

    def __checkControlDir(self):
        """
        Check for the presence of stop_agent file to stop execution of the corresponding Agent
        """
        for agentName in self.__agentModules:
            if not self.__agentModules[agentName]["running"]:
                continue
            agent = self.__agentModules[agentName]["instanceObj"]

            alive = agent.am_getModuleParam("alive")
            if alive:
                if agent.am_checkStopAgentFile():
                    gLogger.info("Found StopAgent file for agent %s" % agentName)
                    alive = False

            if not alive:
                gLogger.info("Stopping agent module %s" % (agentName))
                self.__scheduler.removeTask(self.__agentModules[agentName]["taskId"])
                del self.__tasks[self.__agentModules[agentName]["taskId"]]
                self.__agentModules[agentName]["running"] = False
                agent.am_removeStopAgentFile()



""" DIRAC FileCatalog mix-in class to manage users and groups from the CS Registry
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time
import threading

from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.UserGroupManager.UserAndGroupManagerBase import (
    UserAndGroupManagerBase,
)


class UserAndGroupManagerCS(UserAndGroupManagerBase):
    def getUserAndGroupID(self, credDict):
        user = credDict.get("username", "anon")
        group = credDict.get("group", "anon")
        return S_OK((user, group))

    #####################################################################
    #
    #  User related methods
    #
    #####################################################################

    def addUser(self, name):
        return S_OK(name)

    def deleteUser(self, name, force=True):
        return S_OK()

    def getUsers(self):
        res = gConfig.getSections("/Registry/Users")
        if not res["OK"]:
            return res
        userDict = {}
        for user in res["Value"]:
            userDict[user] = user
        return S_OK(userDict)

    def getUserName(self, uid):
        return S_OK(uid)

    def findUser(self, user):
        return S_OK(user)

    #####################################################################
    #
    #  Group related methods
    #
    #####################################################################

    def addGroup(self, gname):
        return S_OK(gname)

    def deleteGroup(self, gname, force=True):
        return S_OK()

    def getGroups(self):
        res = gConfig.getSections("/Registry/Groups")
        if not res["OK"]:
            return res
        groupDict = {}
        for group in res["Value"]:
            groupDict[group] = group
        return S_OK(groupDict)

    def getGroupName(self, gid):
        return S_OK(gid)

    def findGroup(self, group):
        return S_OK(group)



""" DIRAC FileCatalog component representing a directory tree with
    a closure table

    General warning: when we return the number of affected row, if the values did not change
                     then they are not taken into account, so we might return "Dir does not exist"
                     while it does.... the timestamp update should prevent this to happen, however if
                     you do it several times within 1 second, then there will be no changed, and affected = 0

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import os

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.List import intListToString, stringListToString
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectoryManager.DirectoryTreeBase import DirectoryTreeBase

__RCSID__ = "$Id$"


class DirectoryClosure(DirectoryTreeBase):
    """Class managing Directory Tree with a closure table
    http://technobytz.com/closure_table_store_hierarchical_data.html
    http://fungus.teststation.com/~jon/treehandling/TreeHandling.htm
    http://www.slideshare.net/billkarwin/sql-antipatterns-strike-back
    http://dirtsimple.org/2010/11/simplest-way-to-do-tree-based-queries.html
    """

    def __init__(self, database=None):
        DirectoryTreeBase.__init__(self, database)
        self.directoryTable = "FC_DirectoryList"
        self.closureTable = "FC_DirectoryClosure"

    def findDir(self, path, connection=False):
        """Find directory ID for the given path

        :param path: path of the directory

        :returns: S_OK(id) and res['Level'] as the depth
        """

        dpath = os.path.normpath(path)
        result = self.db.executeStoredProcedure("ps_find_dir", (dpath, "ret1", "ret2"), outputIds=[1, 2])
        if not result["OK"]:
            return result

        if not result["Value"]:
            return S_OK(0)

        res = S_OK(result["Value"][0])
        res["Level"] = result["Value"][1]
        return res

    def findDirs(self, paths, connection=False):
        """Find DirIDs for the given path list

        :param paths: list of path

        :returns: S_OK( { path : ID} )
        """

        dirDict = {}
        if not paths:
            return S_OK(dirDict)
        dpaths = stringListToString([os.path.normpath(path) for path in paths])
        result = self.db.executeStoredProcedureWithCursor("ps_find_dirs", (dpaths,))
        if not result["OK"]:
            return result
        for dirName, dirID in result["Value"]:
            dirDict[dirName] = dirID

        return S_OK(dirDict)

    def removeDir(self, path):
        """Remove directory

        Removing a non existing directory is successful. In that case, DirID is 0

        :param path: path of the dir

        :returns: S_OK() and res['DirID'] the id of the directory removed
        """

        # Find the directory ID
        result = self.findDir(path)
        if not result["OK"]:
            return result

        # If the directory does not exist, we exit successfully with DirID = 0
        if not result["Value"]:
            res = S_OK()
            res["DirID"] = 0
            return res

        dirId = result["Value"]
        result = self.db.executeStoredProcedure("ps_remove_dir", (dirId,), outputIds=[])
        if not result["OK"]:
            return result

        result["DirID"] = dirId
        return result

    def existsDir(self, path):
        """Check the existence of a directory at the specified path

        :param path: directory path

        :returns: S_OK( { 'Exists' : False } ) if the directory does not exist
                  S_OK( { 'Exists' : True, 'DirID' : directory id  } ) if the directory exists
        """

        result = self.findDir(path)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK({"Exists": False})
        else:
            return S_OK({"Exists": True, "DirID": result["Value"]})

    def getDirectoryPath(self, dirID):
        """Get directory name by directory ID

        :param dirID: directory ID

        :returns: S_OK(dir name), or S_ERROR if it does not exist

        """

        result = self.db.executeStoredProcedure("ps_get_dirName_from_id", (dirID, "out"), outputIds=[1])
        if not result["OK"]:
            return result

        dirName = result["Value"][0]

        if not dirName:
            return S_ERROR("Directory with id %d not found" % int(dirID))

        return S_OK(dirName)

    def getDirectoryPaths(self, dirIDList):
        """Get directory names by directory ID list

        :param dirIDList: list of dirIds
        :returns: S_OK( { dirID : dirName} )
        """

        dirs = dirIDList
        if not isinstance(dirIDList, list):
            dirs = [dirIDList]

        dirDict = {}

        # Format the list
        dIds = intListToString(dirs)
        result = self.db.executeStoredProcedureWithCursor("ps_get_dirNames_from_ids", (dIds,))
        if not result["OK"]:
            return result

        for dirId, dirName in result["Value"]:
            dirDict[dirId] = dirName

        return S_OK(dirDict)

    def getPathIDs(self, path):
        """Get IDs of all the directories in the parent hierarchy for a directory
        specified by its path, including itself

        :param path: path of the directory

        :returns: S_OK( list of ids ), S_ERROR if not found
        """

        result = self.findDir(path)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Directory %s not found" % path)

        dirID = result["Value"]

        return self.getPathIDsByID(dirID)

    def getPathIDsByID(self, dirID):
        """Get IDs of all the directories in the parent hierarchy for a directory
        specified by its ID, including itself

        :param dirID: id of the dictionary

        :returns: S_OK( list of ids )

        """

        result = self.db.executeStoredProcedureWithCursor("ps_get_parentIds_from_id", (dirID,))

        if not result["OK"]:
            return result

        return S_OK([dId[0] for dId in result["Value"]])

    def getChildren(self, path, connection=False):
        """Get child directory IDs for the given directory"""
        if isinstance(path, six.string_types):
            result = self.findDir(path, connection)
            if not result["OK"]:
                return result
            if not result["Value"]:
                return S_ERROR("Directory does not exist: %s" % path)
            dirID = result["Value"]
        else:
            dirID = path

        result = self.db.executeStoredProcedureWithCursor("ps_get_direct_children", (dirID,))
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK([])

        return S_OK([x[0] for x in result["Value"]])

    def getSubdirectoriesByID(self, dirID, requestString=False, includeParent=False):
        """Get all the subdirectories of the given directory at a given level

        :param dirID: id of the directory
        :param requestString: if true, returns an sql query to get the information
        :param includeParent: if true, the parent (dirID) will be included

        :returns: S_OK ( { dirID, depth } ) if requestString is False
                 S_OK(request) if requestString is True

        """

        if requestString:
            reqStr = "SELECT ChildID FROM FC_DirectoryClosure WHERE ParentID = %s" % dirID
            if not includeParent:
                reqStr += " AND Depth != 0"
            return S_OK(reqStr)

        result = self.db.executeStoredProcedureWithCursor("ps_get_sub_directories", (dirID, includeParent))
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK({})

        return S_OK(dict((x[0], x[1]) for x in result["Value"]))

    def getAllSubdirectoriesByID(self, dirIdList):
        """Get IDs of all the subdirectories of directories in a given list

        :param dirList: list of dir Ids
        :returns: S_OK([ unordered dir ids ])
        """

        dirs = dirIdList
        if not isinstance(dirIdList, list):
            dirs = [dirIdList]

        dIds = intListToString(dirs)
        result = self.db.executeStoredProcedureWithCursor("ps_get_multiple_sub_directories", (dIds,))

        if not result["OK"]:
            return result

        resultList = [dirId[0] for dirId in result["Value"]]
        return S_OK(resultList)

    def getSubdirectories(self, path):
        """Get subdirectories of the given directory

        :param path: path of the directory

        :returns: S_OK ( { dirID, depth } )
        """

        result = self.findDir(path)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK({})

        dirID = result["Value"]
        result = self.getSubdirectoriesByID(dirID)
        return result

    def countSubdirectories(self, dirId, includeParent=True):
        """Count the number of subdirectories

        :param dirID: id of the directory
        :param includeParent: count itself

        :returns: S_OK(value)
        """

        result = self.db.executeStoredProcedure(
            "ps_count_sub_directories", (dirId, includeParent, "ret1"), outputIds=[2]
        )
        if not result["OK"]:
            return result

        res = S_OK(result["Value"][0])
        return res

    ########################################################################################################
    #
    #  We overwrite some methods from the base class because of the new DB constraints or perf reasons
    #
    #  Some methods could be inherited in the future if we have perf problems. For example
    #  * removeDirectory
    #  * changeDirectory[Group/Owner/Mode]
    #  * getDirectoryPermissions (when called by getPathPermissions, we could buffer the getUserAndGroupID call)
    #  * getFileIDsInDirectory (used only by DirectoryMetadata)
    #  * getFilesInDirectory (used only by DirectoryMetadata)
    #  * getFileLFNsInDirectory (used only by FileMetadata)
    #  * getFileLFNsInDirectoryByDirectory (used only by FileMetadata)
    #  * _getDirectoryContents (we could bring together some requests)
    #
    ########################################################################################################

    def makeDirectory(self, path, credDict, status=1):
        """Create a directory

        :param path: has to be an absolute path. The parent dir has to exist
        :param credDict: credential dict of the owner of the directory
        :param: status ????

        :returns: S_OK (dirID) with a flag res['NewDirectory'] to True or False
                S_ERROR if there is a problem, or if there is no parent
        """

        if path[0] != "/":
            return S_ERROR("Not an absolute path")

        # Strip off the trailing slash if necessary
        dpath = os.path.normpath(path)
        parentDir = os.path.dirname(dpath)

        # Try to see if the dir exists
        result = self.findDir(path)
        if not result["OK"]:
            return result

        # if it does, we return it's id, with a flag NewDirectory to false
        dirID = result["Value"]
        if dirID:
            result = S_OK(dirID)
            result["NewDirectory"] = False
            return result

        # If it is the root directory, we force the owner to 'root'/'root' (id 1 in the db)
        if path == "/":
            l_uid = 1
            l_gid = 1
        else:
            # get the uid/gid of the owner
            result = self.db.ugManager.getUserAndGroupID(credDict)
            if not result["OK"]:
                return result
            (l_uid, l_gid) = result["Value"]

        # Find the ID of the parent
        res = self.findDir(parentDir)
        if not res["OK"]:
            return res

        parentDirId = res["Value"]

        # We only insert if there is a parent or if it is the root '/'
        if parentDirId or path == "/":
            result = self.db.executeStoredProcedureWithCursor(
                packageName="ps_insert_dir", parameters=(parentDirId, dpath, l_uid, l_gid, self.db.umask, status)
            )

            if not result["OK"]:
                return result

            dirId = result["Value"][0][0]

            result = S_OK(dirId)
            result["NewDirectory"] = True
            return result
        else:
            return S_ERROR("Cannot create directory without parent")

    def isEmpty(self, path):
        """Find out if the given directory is empty

        Rem: the speed could be enhanced if we were joining the FC_Files and FC_Directory* in the query.
            For the time being, it can stay like this

        :param path: path of the directory

        :returns: S_OK(true) if there are no file nor directorie, S_OK(False) otherwise
        """

        result = self.findDir(path)
        if not result["OK"]:
            return result
        dirId = result["Value"]

        if not dirId:
            return S_ERROR("Directory does not exist %s" % path)

        # Check if there are subdirectories
        result = self.countSubdirectories(dirId, includeParent=False)
        if not result["OK"]:
            return result

        subDirCount = result["Value"]
        if subDirCount:
            return S_OK(False)

        # Check if there are files in it
        result = self.db.fileManager.countFilesInDir(dirId)
        if not result["OK"]:
            return result

        fileCount = result["Value"]

        if fileCount:
            return S_OK(False)

        # If no files or subdir, it's empty
        return S_OK(True)

    def getDirectoryParameters(self, pathOrDirId):
        """Get parameters of the given directory

        :param pathOrDirID: the path or the id of the directory

        :returns: S_OK(dict), where dict has the following keys:
                        "DirID", "UID", "Owner", "GID", "OwnerGroup", "Status", "Mode", "CreationDate", "ModificationDate"
        """
        # Which procedure to use
        psName = None
        # it is a path ...
        if isinstance(pathOrDirId, six.string_types):
            psName = "ps_get_all_directory_info"
        # it is the dirId
        elif isinstance(pathOrDirId, ((list,) + six.integer_types)):
            psName = "ps_get_all_directory_info_from_id"
        else:
            return S_ERROR("Unknown type of pathOrDirId %s" % type(pathOrDirId))

        result = self.db.executeStoredProcedureWithCursor(psName, (pathOrDirId,))
        if not result["OK"]:
            return result

        # All the fields returned
        fieldNames = [
            "DirID",
            "UID",
            "Owner",
            "GID",
            "OwnerGroup",
            "Status",
            "Mode",
            "CreationDate",
            "ModificationDate",
        ]

        if not result["Value"]:
            return S_ERROR("Directory does not exist %s" % pathOrDirId)

        row = result["Value"][0]

        # Create a dictionary from the fieldNames
        rowDict = dict(zip(fieldNames, row))

        return S_OK(rowDict)

    def _setDirectoryParameter(self, path, pname, pvalue, recursive=False):
        """Set a numerical directory parameter


        Rem: the parent class has a more generic method, which is called
             in case we are given an unknown parameter

        :param path: path of the directory
        :param pname: name of the parameter to set
        :param pvalue: value of the parameter (an id or a value)

        :returns: S_OK(nb of row changed). It should always be 1 !
                S_ERROR if the directory does not exist
        """

        # The PS associated with a given parameter
        psNames = {
            "UID": "ps_set_dir_uid",
            "GID": "ps_set_dir_gid",
            "Status": "ps_set_dir_status",
            "Mode": "ps_set_dir_mode",
        }

        psName = psNames.get(pname, None)

        # If we have a recursive procedure and it is wanted, call it
        if recursive and pname in ["UID", "GID", "Mode"] and psName:
            psName += "_recursive"

        # If there is an associated procedure, we go for it
        if psName:
            result = self.db.executeStoredProcedureWithCursor(psName, (path, pvalue))

            if not result["OK"]:
                return result

            errno, affected, errMsg = result["Value"][0]
            if errno:
                return S_ERROR(errMsg)

            if not affected:
                # Either there were no changes, or the directory does not exist
                exists = self.existsDir(path).get("Value", {}).get("Exists")
                if not exists:
                    return S_ERROR(errno.ENOENT, "Directory does not exist: %s" % path)
                affected = 1

            return S_OK(affected)

        # In case this is a 'new' parameter, we have a fallback solution, but we should add a specific ps for it
        else:
            return DirectoryTreeBase._setDirectoryParameter(self, path, pname, pvalue)

    def _setDirectoryGroup(self, path, gname, recursive=False):
        """Set the directory owner"""

        result = self.db.ugManager.findGroup(gname)
        if not result["OK"]:
            return result

        gid = result["Value"]

        return self._setDirectoryParameter(path, "GID", gid, recursive=recursive)

    def _setDirectoryOwner(self, path, owner, recursive=False):
        """Set the directory owner"""

        result = self.db.ugManager.findUser(owner)
        if not result["OK"]:
            return result

        uid = result["Value"]

        return self._setDirectoryParameter(path, "UID", uid, recursive=recursive)

    def _setDirectoryMode(self, path, mode, recursive=False):
        """set the directory mode

        :param mixed path: directory path as a string or int or list of ints or select statement
        :param int mode: new mode
        """
        return self._setDirectoryParameter(path, "Mode", mode, recursive=recursive)

    def __getLogicalSize(self, lfns, ps_name, recursiveSum=True, connection=None):
        successful = {}
        failed = {}
        for path in lfns:
            result = self.findDir(path)
            if not result["OK"] or not result["Value"]:
                failed[path] = "Directory not found"
                continue

            dirID = result["Value"]
            result = self.db.executeStoredProcedureWithCursor(ps_name, (dirID, recursiveSum))

            if not result["OK"]:
                failed[path] = result["Message"]

            elif result["Value"]:
                successful[path] = {
                    "LogicalSize": int(result["Value"][0][0]),
                    "LogicalFiles": int(result["Value"][0][1]),
                }

                result = self.countSubdirectories(dirID, includeParent=False)
                if result["OK"]:
                    successful[path]["LogicalDirectories"] = result["Value"]
                else:
                    successful[path]["LogicalDirectories"] = -1

            else:
                successful[path] = {"LogicalSize": 0, "LogicalFiles": 0, "LogicalDirectories": 0}

        return S_OK({"Successful": successful, "Failed": failed})

    def _getDirectoryLogicalSizeFromUsage(self, lfns, recursiveSum=True, connection=None):
        """Get the total "logical" size of the requested directories"""
        return self.__getLogicalSize(lfns, "ps_get_dir_logical_size", recursiveSum=recursiveSum, connection=connection)

    def _getDirectoryLogicalSize(self, lfns, recursiveSum=True, connection=None):
        """Get the total "logical" size of the requested directories"""
        return self.__getLogicalSize(
            lfns, "ps_calculate_dir_logical_size", recursiveSum=recursiveSum, connection=connection
        )

    def __getPhysicalSize(self, lfns, ps_name, recursiveSum=True, connection=None):
        """Get the total size of the requested directories"""

        successful = {}
        failed = {}
        for path in lfns:
            result = self.findDir(path)
            if not result["OK"]:
                failed[path] = "Directory not found"
                continue
            if not result["Value"]:
                failed[path] = "Directory not found"
                continue
            dirID = result["Value"]

            result = self.db.executeStoredProcedureWithCursor(ps_name, (dirID, recursiveSum))
            if not result["OK"]:
                failed[path] = result["Message"]
                continue

            if result["Value"]:
                seDict = {}
                totalSize = 0
                totalFiles = 0
                for seName, seSize, seFiles in result["Value"]:
                    seDict[seName] = {"Size": int(seSize), "Files": int(seFiles)}
                    totalSize += seSize
                    totalFiles += seFiles
                seDict["TotalSize"] = int(totalSize)
                seDict["TotalFiles"] = int(totalFiles)
                successful[path] = seDict

            else:
                successful[path] = {}

        return S_OK({"Successful": successful, "Failed": failed})

    def _getDirectoryPhysicalSizeFromUsage(self, lfns, recursiveSum=True, connection=None):
        """Get the total size of the requested directories"""
        return self.__getPhysicalSize(
            lfns, "ps_get_dir_physical_size", recursiveSum=recursiveSum, connection=connection
        )

    def _getDirectoryPhysicalSize(self, lfns, recursiveSum=True, connection=None):
        """Get the total size of the requested directories"""
        return self.__getPhysicalSize(
            lfns, "ps_calculate_dir_physical_size", recursiveSum=recursiveSum, connection=None
        )

    def _changeDirectoryParameter(self, paths, directoryFunction, _fileFunction, recursive=False):
        """Bulk setting of the directory parameter with recursion for all the subdirectories and files

        :param dict paths: dictionary < lfn : value >, where value is the value of parameter to be set
        :param function directoryFunction: function to change directory(ies) parameter
        :param function fileFunction: function to change file(s) parameter
        :param bool recursive: flag to apply the operation recursively
        """

        arguments = paths
        successful = {}
        failed = {}
        for path, attribute in arguments.items():
            result = directoryFunction(path, attribute, recursive=recursive)
            if not result["OK"]:
                failed[path] = result["Message"]
            else:
                successful[path] = True

        return S_OK({"Successful": successful, "Failed": failed})
